import random

from aiohttp_retry import (
    ExponentialRetry,
    FibonacciRetry,
    JitterRetry,
    ListRetry,
    RandomRetry,
)


def test_exponential_retry():
    retry = ExponentialRetry(attempts=10)
    timeouts = [retry.get_timeout(x) for x in range(10)]
    assert timeouts == [0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, 25.6, 30.0]


def test_random_retry():
    retry = RandomRetry(attempts=10, random_func=random.Random(0).random)
    timeouts = [round(retry.get_timeout(x), 2) for x in range(10)]
    assert timeouts == [2.55, 2.3, 1.32, 0.85, 1.58, 1.27, 2.37, 0.98, 1.48, 1.79]


def test_list_retry():
    expected = [1.2, 2.1, 3.4, 4.3, 4.5, 5.4, 5.6, 6.5, 6.7, 7.6]
    retry = ListRetry(expected)
    timeouts = [retry.get_timeout(x) for x in range(10)]
    assert timeouts == expected


def test_fibonacci_retry():
    retry = FibonacciRetry(attempts=10, multiplier=2, max_timeout=60)
    timeouts = [retry.get_timeout(x) for x in range(10)]
    assert timeouts == [4.0, 6.0, 10.0, 16.0, 26.0, 42.0, 60, 60, 60, 60]


def test_jitter_retry():
    random.seed(10)
    retry = JitterRetry(attempts=10)
    timeouts = [retry.get_timeout(x) for x in range(10)]
    assert len(timeouts) == 10

    expected = [
        1.4,
        0.9,
        1.7,
        0.9,
        4.2,
        5.9,
        8.1,
        12.9,
        26.6,
        30.4,
    ]
    for idx, timeout in enumerate(timeouts):
        assert abs(timeout - expected[idx]) < 0.1



from types import SimpleNamespace
from typing import Optional, Tuple

import pytest
from aiohttp import (
    BasicAuth,
    ClientResponse,
    ClientResponseError,
    ClientSession,
    TraceConfig,
    TraceRequestStartParams,
    hdrs,
)
from yarl import URL

from aiohttp_retry import ExponentialRetry, ListRetry, RetryClient
from aiohttp_retry.client import RequestParams
from aiohttp_retry.retry_options import RetryOptionsBase
from tests.app import App


async def get_retry_client_and_test_app_for_test(
    aiohttp_client,
    raise_for_status: bool = False,
    retry_options: Optional[RetryOptionsBase] = None,
) -> Tuple[RetryClient, App]:
    test_app = App()
    app = test_app.web_app()
    client = await aiohttp_client(app, raise_for_status=raise_for_status)

    retry_client = RetryClient(client_session=client, retry_options=retry_options)
    return retry_client, test_app


async def test_hello(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.get('/ping') as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 1

    await retry_client.close()


async def test_hello_by_request(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.request(method=hdrs.METH_GET, url='/ping') as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 1

    await retry_client.close()


async def test_hello_with_context(aiohttp_client):
    test_app = App()
    app = test_app.web_app()
    client = await aiohttp_client(app)
    async with RetryClient() as retry_client:
        retry_client._client = client
        async with retry_client.get('/ping') as response:
            text = await response.text()
            assert response.status == 200
            assert text == 'Ok!'

            assert test_app.counter == 1


async def test_internal_error(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    retry_options = ExponentialRetry(attempts=5)
    async with retry_client.get('/internal_error', retry_options) as response:
        assert response.status == 500
        assert test_app.counter == 5

    await retry_client.close()


async def test_not_found_error(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    retry_options = ExponentialRetry(attempts=5, statuses={404})
    async with retry_client.get('/not_found_error', retry_options) as response:
        assert response.status == 404
        assert test_app.counter == 5

    await retry_client.close()


async def test_sometimes_error(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    retry_options = ExponentialRetry(attempts=5)
    async with retry_client.get('/sometimes_error', retry_options) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 3

    await retry_client.close()


async def test_sometimes_error_with_raise_for_status(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client, raise_for_status=True)
    retry_options = ExponentialRetry(attempts=5, exceptions={ClientResponseError})
    async with retry_client.get('/sometimes_error', retry_options) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 3

    await retry_client.close()


async def test_override_options(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(
        aiohttp_client,
        retry_options=ExponentialRetry(attempts=1)
    )
    retry_options = ExponentialRetry(attempts=5)
    async with retry_client.get('/sometimes_error', retry_options) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 3

    await retry_client.close()


async def test_hello_awaitable(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    response = await retry_client.get('/ping')
    text = await response.text()
    assert response.status == 200
    assert text == 'Ok!'

    assert test_app.counter == 1

    await retry_client.close()


async def test_add_trace_request_ctx(aiohttp_client):
    actual_request_contexts = []

    async def on_request_start(
        _: ClientSession,
        trace_config_ctx: SimpleNamespace,
        __: TraceRequestStartParams,
    ) -> None:
        actual_request_contexts.append(trace_config_ctx)

    test_app = App()

    trace_config = TraceConfig()
    trace_config.on_request_start.append(on_request_start)  # type: ignore

    retry_client = RetryClient()
    retry_client._client = await aiohttp_client(
        test_app.web_app(),
        trace_configs=[trace_config]
    )

    async with retry_client.get('/sometimes_error', trace_request_ctx={'foo': 'bar'}):
        assert test_app.counter == 3

    assert actual_request_contexts == [
        SimpleNamespace(
            trace_request_ctx={
                'foo': 'bar',
                'current_attempt': i + 1,
            },
        )
        for i in range(3)
    ]


@pytest.mark.parametrize("attempts", [2, 3])
async def test_change_urls_in_request(aiohttp_client, attempts):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(
        aiohttp_client,
        retry_options=ExponentialRetry(attempts=attempts)
    )
    async with retry_client.get(url=['/internal_error', '/ping']) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 2

    await retry_client.close()


@pytest.mark.parametrize("attempts", [2, 3])
async def test_change_urls_as_tuple_in_request(aiohttp_client, attempts):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(
        aiohttp_client,
        retry_options=ExponentialRetry(attempts=attempts)
    )
    async with retry_client.get(url=('/internal_error', '/ping')) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 2

    await retry_client.close()


@pytest.mark.parametrize("url", [{"/ping", "/internal_error"}, []])
async def test_pass_bad_urls(aiohttp_client, url):
    retry_client, _ = await get_retry_client_and_test_app_for_test(aiohttp_client)
    with pytest.raises(ValueError):
        async with retry_client.get(url=url):
            pass

    await retry_client.close()


@pytest.mark.parametrize("url, method", [
    ("/options_handler", 'options'),
    ("/head_handler", 'head'),
    ("/post_handler", 'post'),
    ("/put_handler", 'put'),
    ("/patch_handler", 'patch'),
    ("/delete_handler", 'delete'),
])
async def test_methods(aiohttp_client, url, method):
    retry_client, _ = await get_retry_client_and_test_app_for_test(aiohttp_client)
    method_func = getattr(retry_client, method)
    async with method_func(url) as response:
        assert response.method.lower() == method

    await retry_client.close()


async def test_not_found_error_with_retry_client_raise_for_status(aiohttp_client):
    test_app = App()
    app = test_app.web_app

    client = await aiohttp_client(app)
    retry_client = RetryClient(raise_for_status=True)
    retry_client._client = client

    retry_options = ExponentialRetry(attempts=5, statuses={404})
    override_response = retry_client.get('/not_found_error', retry_options, raise_for_status=False)
    assert not override_response._raise_for_status
    response = retry_client.get('/not_found_error', retry_options)
    assert response._raise_for_status

    try:
        async with response:
            pass
    except ClientResponseError as exc:
        assert exc.status == 404
        assert test_app.counter == 5
    else:
        raise AssertionError('Expected ClientResponseError not raised')

    await retry_client.close()
    await client.close()


async def test_request(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.request(hdrs.METH_GET, '/ping') as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 1

    await retry_client.close()


async def test_url_as_yarl(aiohttp_client):
    """https://github.com/inyutin/aiohttp_retry/issues/41"""
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.get(URL('/ping')) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 1

    await retry_client.close()


async def test_change_client_retry_options(aiohttp_client):
    retry_options = ExponentialRetry(attempts=5)
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client, retry_options=retry_options)

    # first time with 5 attempts is okay
    async with retry_client.get('/sometimes_error') as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 3

    test_app.counter = 0
    retry_client.retry_options.attempts = 2

    # second time with 5 attempts is error
    async with retry_client.get('/sometimes_error') as response:
        text = await response.text()
        assert response.status == 500
        assert test_app.counter == 2

    await retry_client.close()


async def test_not_retry_server_errors(aiohttp_client):
    retry_options = ExponentialRetry(retry_all_server_errors=False)
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.get('/internal_error', retry_options) as response:
        assert response.status == 500
        assert test_app.counter == 1

    await retry_client.close()


async def test_list_retry_works_for_multiple_attempts(aiohttp_client):
    retry_options = ListRetry(timeouts=[0]*3)
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)

    async with retry_client.get('/internal_error', retry_options) as response:
        assert response.status == 500
        assert test_app.counter == 3

    await retry_client.close()


async def test_implicit_client(aiohttp_client):
    # check that if client not passed that it created implicitly
    test_app = App()

    retry_client = RetryClient()
    assert retry_client._client is not None

    retry_client._client = await aiohttp_client(test_app.web_app())
    async with retry_client.get('/ping') as response:
        assert response.status == 200

    await retry_client.close()


async def test_evaluate_response_callback(aiohttp_client):
    async def evaluate_response(response: ClientResponse) -> bool:
        try:
            await response.json()
        except:
            return False
        return True

    retry_options = ExponentialRetry(attempts=5, evaluate_response_callback=evaluate_response)
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client, retry_options=retry_options)

    async with retry_client.get('/sometimes_json') as response:
        body = await response.json()
        assert response.status == 200
        assert body == {'status': 'Ok!'}

        assert test_app.counter == 3


async def test_multiply_urls_by_requests(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.requests(
        params_list=[
            RequestParams(
                method='GET',
                url='/internal_error',
            ),
            RequestParams(
                method='GET',
                url='/ping',
            ),
        ]
    ) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 2

    await retry_client.close()


async def test_multiply_methods_by_requests(aiohttp_client):
    retry_options = ExponentialRetry(statuses={405})  # method not allowed
    retry_client, _ = await get_retry_client_and_test_app_for_test(aiohttp_client, retry_options=retry_options)
    async with retry_client.requests(
        params_list=[
            RequestParams(
                method='POST',
                url='/ping',
            ),
            RequestParams(
                method='GET',
                url='/ping',
            ),
        ]
    ) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

    await retry_client.close()


async def test_change_headers(aiohttp_client):
    retry_options = ExponentialRetry(statuses={406})
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client, retry_options=retry_options)
    async with retry_client.requests(
        params_list=[
            RequestParams(
                method='GET',
                url='/check_headers',
            ),
            RequestParams(
                method='GET',
                url='/check_headers',
                headers={'correct_headers': 'True'},
            ),
        ]
    ) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 2

    await retry_client.close()


async def test_additional_params(aiohttp_client):
    # https://github.com/inyutin/aiohttp_retry/issues/79
    auth = BasicAuth("username", "password")
    retry_client, _ = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.request(hdrs.METH_GET, '/with_auth', auth=auth) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

    await retry_client.close()


async def test_request_headers(aiohttp_client):
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)
    async with retry_client.get(url='/check_headers', headers={'correct_headers': 'True'}) as response:
        text = await response.text()
        assert response.status == 200
        assert text == 'Ok!'

        assert test_app.counter == 1

    await retry_client.close()


async def test_list_retry_all_failed(aiohttp_client):
    # there was a specific bug
    async def evaluate_response(response: ClientResponse) -> bool:
        return False

    retry_options = ListRetry(timeouts=[1]*3, statuses={403}, evaluate_response_callback=evaluate_response)
    retry_client, test_app = await get_retry_client_and_test_app_for_test(aiohttp_client)

    async with retry_client.get('/with_auth', retry_options=retry_options) as response:
        assert response.status == 403
        assert test_app.counter == 3

    await retry_client.close()



from aiohttp import web


class App:
    def __init__(self):
        self.counter = 0

        app = web.Application()
        app.router.add_get('/ping', self.ping_handler)
        app.router.add_get('/internal_error', self.internal_error_handler)
        app.router.add_get('/not_found_error', self.not_found_error_handler)
        app.router.add_get('/sometimes_error', self.sometimes_error)
        app.router.add_get('/sometimes_json', self.sometimes_json)
        app.router.add_get('/check_headers', self.check_headers)
        app.router.add_get('/with_auth', self.with_auth)

        app.router.add_options('/options_handler', self.ping_handler)
        app.router.add_head('/head_handler', self.ping_handler)
        app.router.add_post('/post_handler', self.ping_handler)
        app.router.add_put('/put_handler', self.ping_handler)
        app.router.add_patch('/patch_handler', self.ping_handler)
        app.router.add_delete('/delete_handler', self.ping_handler)

        self._web_app = app

    async def ping_handler(self, _: web.Request) -> web.Response:
        self.counter += 1
        return web.Response(text='Ok!', status=200)

    async def internal_error_handler(self, _: web.Request) -> web.Response:
        self.counter += 1
        raise web.HTTPInternalServerError()

    async def not_found_error_handler(self, _: web.Request) -> web.Response:
        self.counter += 1
        raise web.HTTPNotFound()

    async def sometimes_error(self, _: web.Request) -> web.Response:
        self.counter += 1
        if self.counter == 3:
            return web.Response(text='Ok!', status=200)

        raise web.HTTPInternalServerError()

    async def sometimes_json(self, _: web.Request) -> web.Response:
        self.counter += 1
        if self.counter == 3:
            return web.json_response(data={'status': 'Ok!'}, status=200)

        return web.Response(text='Ok!', status=200)

    async def check_headers(self, request: web.Request) -> web.Response:
        self.counter += 1
        if request.headers.get('correct_headers') != 'True':
            raise web.HTTPNotAcceptable()

        return web.Response(text='Ok!', status=200)

    async def with_auth(self, request: web.Request) -> web.Response:
        self.counter += 1

        # BasicAuth("username", "password")
        if request.headers.get('Authorization') != 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=':
            return web.Response(text='incorrect auth', status=403)
        return web.Response(text='Ok!', status=200)

    @property
    def web_app(self) -> web.Application:
        return self._web_app



from setuptools import find_packages, setup


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name='aiohttp_retry',
    version='2.8.3',
    description='Simple retry client for aiohttp',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='aiohttp retry client',
    author='Dmitry Inyutin',
    author_email='inyutin.da@gmail.com',
    url='https://github.com/inyutin/aiohttp_retry',
    license='MIT',
    include_package_data=True,
    packages=find_packages(exclude=["tests", "tests.*"]),
    platforms=['any'],
    python_requires='>=3.7',
    install_requires=[
        'aiohttp',
    ],
    package_data={
        "aiohttp_retry": ["py.typed"],
    }
)
